commit 82ebcf26b67cee47a54c5474a2d57eb64d0d14f4
Author: Alex Wilson <alex@cooperi.net>
Date:   Wed Apr 9 16:24:36 2014 +1000

    wip: tls support for heartbeat

diff --git a/lib/ssl/src/ssl_connection.erl b/lib/ssl/src/ssl_connection.erl
index 8210693..0aa4df9 100644
--- a/lib/ssl/src/ssl_connection.erl
+++ b/lib/ssl/src/ssl_connection.erl
@@ -27,6 +27,7 @@
 
 -include("ssl_api.hrl").
 -include("ssl_connection.hrl").
+-include("tls_handshake.hrl").
 -include("ssl_handshake.hrl").
 -include("ssl_alert.hrl").
 -include("ssl_record.hrl").
@@ -45,6 +46,8 @@
 	 peer_certificate/1, renegotiation/1, negotiated_next_protocol/1, prf/5
 	]).
 
+-export([peer_hello/1, heartbeat/3, heartbeat/4]).
+
 -export([handle_session/6]).
 
 %% SSL FSM state functions
@@ -218,6 +221,14 @@ peer_certificate(ConnectionPid) ->
 renegotiation(ConnectionPid) ->
     sync_send_all_state_event(ConnectionPid, renegotiate).
 
+peer_hello(ConnectionPid) ->
+    sync_send_all_state_event(ConnectionPid, peer_hello).
+
+heartbeat(Pid, Length, Payload) -> heartbeat(Pid, Length, Payload, 5000).
+heartbeat(Pid, Length, Payload, Timeout) when is_integer(Length) and is_binary(Payload) ->
+    gen_fsm:sync_send_all_state_event(Pid, {heartbeat, Length, Payload, Timeout}, Timeout*2).
+
+
 %%--------------------------------------------------------------------
 -spec prf(pid(), binary() | 'master_secret', binary(),
 	  binary() | ssl:prf_random(), non_neg_integer()) ->
@@ -800,7 +811,36 @@ handle_sync_event(session_info, _, StateName,
 handle_sync_event(peer_certificate, _, StateName,
 		  #state{session = #session{peer_certificate = Cert}}
 		  = State) ->
-    {reply, {ok, Cert}, StateName, State, get_timeout(State)}.
+    {reply, {ok, Cert}, StateName, State, get_timeout(State)};
+
+handle_sync_event(peer_hello, _, StateName,
+		 #state{peer_hello = Hello} = State) ->
+    {reply, {ok, Hello}, StateName, State, get_timeout(State)};
+
+handle_sync_event({heartbeat, Len, Payload, Timeout}, From, StateName,
+		 #state{peer_hello = H, protocol_cb = Connection} = State) ->
+    Ex = case H of
+        #client_hello{extensions = E} -> E;
+        #server_hello{extensions = E} -> E
+    end,
+    case Ex#hello_extensions.heartbeat of
+    	undefined ->
+    	    {reply, {error, no_heartbeat_in_hello},
+    		StateName, State, get_timeout(State)};
+    	#heartbeat{mode = peer_not_allowed_to_send} ->
+    	    {reply, {error, heartbeat_disabled},
+    		StateName, State, get_timeout(State)};
+    	#heartbeat{mode = peer_allowed_to_send} ->
+            if (State#state.heartbeat_from =:= undefined) ->
+    	        {ok, CS2} = Connection:handle_own_heartbeat(request, Len, Payload, State),
+    	        {ok, Ref} = timer:send_after(Timeout, heartbeat_timeout),
+    	        {next_state, StateName,
+                    State#state{heartbeat_from = From, heartbeat_tref = Ref,
+                        connection_states = CS2}, get_timeout(State)};
+            true ->
+                {reply, {error, busy}, StateName, State, get_timeout(State)}
+            end
+    end.
 
 handle_info({ErrorTag, Socket, econnaborted}, StateName,
 	    #state{socket = Socket, transport_cb = Transport,
diff --git a/lib/ssl/src/ssl_connection.hrl b/lib/ssl/src/ssl_connection.hrl
index 27489ca..eb1ef40 100644
--- a/lib/ssl/src/ssl_connection.hrl
+++ b/lib/ssl/src/ssl_connection.hrl
@@ -78,7 +78,9 @@
 	  allow_renegotiate = true                    ::boolean(),
           expecting_next_protocol_negotiation = false ::boolean(),
           next_protocol = undefined                   :: undefined | binary(),
-	  client_ecc          % {Curves, PointFmt}
+	  client_ecc,          % {Curves, PointFmt}
+	  peer_hello,
+	  heartbeat_from, heartbeat_tref
 	 }).
 
 -define(DEFAULT_DIFFIE_HELLMAN_PARAMS,
diff --git a/lib/ssl/src/ssl_handshake.erl b/lib/ssl/src/ssl_handshake.erl
index 2b9bae6..19e26b0 100644
--- a/lib/ssl/src/ssl_handshake.erl
+++ b/lib/ssl/src/ssl_handshake.erl
@@ -55,7 +55,7 @@
 	 select_session/10, supported_ecc/1]).
 
 %% Extensions handling
--export([client_hello_extensions/6,
+-export([client_hello_extensions/7,
 	 handle_client_hello_extensions/9, %% Returns server hello extensions
 	 handle_server_hello_extensions/9, select_curve/2
 	]).
@@ -106,7 +106,7 @@ server_hello(SessionId, Version, ConnectionStates, Extensions) ->
 server_hello_done() ->
     #server_hello_done{}.
 
-client_hello_extensions(Host, Version, CipherSuites, SslOpts, ConnectionStates, Renegotiation) ->
+client_hello_extensions(Host, Version, CipherSuites, SslOpts, HeartbeatMode, ConnectionStates, Renegotiation) ->
     {EcPointFormats, EllipticCurves} =
 	case advertises_ec_ciphers(lists:map(fun ssl_cipher:suite_definition/1, CipherSuites)) of
 	    true ->
@@ -126,6 +126,7 @@ client_hello_extensions(Host, Version, CipherSuites, SslOpts, ConnectionStates,
        next_protocol_negotiation =
 	   encode_client_protocol_negotiation(SslOpts#ssl_options.next_protocol_selector,
 					      Renegotiation),
+	   heartbeat = #heartbeat{mode = HeartbeatMode},
        sni = sni(Host, SslOpts#ssl_options.server_name_indication)}.
 
 %%--------------------------------------------------------------------
@@ -752,6 +753,14 @@ encode_hello_extensions([#renegotiation_info{renegotiated_connection = ?byte(0)
     Len = byte_size(Info),
     encode_hello_extensions(Rest, <<?UINT16(?RENEGOTIATION_EXT), ?UINT16(Len), Info/binary, Acc/binary>>);
 
+encode_hello_extensions([#heartbeat{mode = ModeAtom} | Rest], Acc) ->
+	Mode = case ModeAtom of
+		peer_allowed_to_send -> 1;
+		peer_not_allowed_to_send -> 2
+	end,
+	Info = <<Mode:8>>,
+	encode_hello_extensions(Rest, <<?UINT16(?HEARTBEAT_EXT), ?UINT16(1), Info/binary, Acc/binary>>);
+
 encode_hello_extensions([#renegotiation_info{renegotiated_connection = Info} | Rest], Acc) ->
     InfoLen = byte_size(Info),
     Len = InfoLen +1,
@@ -1109,7 +1118,8 @@ handle_client_hello_extensions(RecordCB, Random, ClientCipherSuites,
 									ConnectionStates, Renegotiation),
 				ec_point_formats = server_ecc_extension(Version, ECCFormat),
 				next_protocol_negotiation =
-				    encode_protocols_advertised_on_server(ProtocolsToAdvertise)
+				    encode_protocols_advertised_on_server(ProtocolsToAdvertise),
+				heartbeat = #heartbeat{mode = peer_allowed_to_send}
 			       },
     {Session, ConnectionStates, ServerHelloExtensions}.
 
@@ -1235,10 +1245,11 @@ hello_extensions_list(#hello_extensions{renegotiation_info = RenegotiationInfo,
 					hash_signs = HashSigns,
 					ec_point_formats = EcPointFormats,
 					elliptic_curves = EllipticCurves,
+					heartbeat = HB,
 					next_protocol_negotiation = NextProtocolNegotiation,
 					sni = Sni}) ->
     [Ext || Ext <- [RenegotiationInfo, SRP, HashSigns,
-		    EcPointFormats, EllipticCurves, NextProtocolNegotiation, Sni], Ext =/= undefined].
+		    EcPointFormats, EllipticCurves, NextProtocolNegotiation, Sni, HB], Ext =/= undefined].
 
 srp_user(#ssl_options{srp_identity = {UserName, _}}) ->
     #srp{username = UserName};
@@ -1650,7 +1661,16 @@ dec_hello_extensions(<<?UINT16(?SIGNATURE_ALGORITHMS_EXT), ?UINT16(Len),
 			<<?BYTE(Hash), ?BYTE(Sign)>> <= SignAlgoList],
     dec_hello_extensions(Rest, Acc#hello_extensions{hash_signs =
 						    #hash_sign_algos{hash_sign_algos = HashSignAlgos}});
-
+dec_hello_extensions(<<?UINT16(?HEARTBEAT_EXT), ?UINT16(Len),
+			   Info:Len/binary, Rest/binary>>, Acc) ->
+	<<Mode:8>> = Info,
+	ModeAtom = case Mode of
+		1 -> peer_allowed_to_send;
+		2 -> peer_not_allowed_to_send;
+		_ -> unknown
+	end,
+	dec_hello_extensions(Rest, Acc#hello_extensions{heartbeat =
+		#heartbeat{mode = ModeAtom}});
 dec_hello_extensions(<<?UINT16(?ELLIPTIC_CURVES_EXT), ?UINT16(Len),
 		       ExtData:Len/binary, Rest/binary>>, Acc) ->
     <<?UINT16(_), EllipticCurveList/binary>> = ExtData,
diff --git a/lib/ssl/src/ssl_handshake.hrl b/lib/ssl/src/ssl_handshake.hrl
index 7516052..34c473b 100644
--- a/lib/ssl/src/ssl_handshake.hrl
+++ b/lib/ssl/src/ssl_handshake.hrl
@@ -96,6 +96,7 @@
 	  renegotiation_info,
 	  hash_signs,          % supported combinations of hashes/signature algos
 	  next_protocol_negotiation = undefined, % [binary()]
+	  heartbeat,
 	  srp,
 	  ec_point_formats,
 	  elliptic_curves,
@@ -300,6 +301,9 @@
 	  hash_sign_algos
 	 }).
 
+-define(HEARTBEAT_EXT, 15).
+-record(heartbeat, {mode}).
+
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %% Next Protocol Negotiation
 %% (http://tools.ietf.org/html/draft-agl-tls-nextprotoneg-02)
diff --git a/lib/ssl/src/ssl_record.erl b/lib/ssl/src/ssl_record.erl
index 018c8be..c43ebb2 100644
--- a/lib/ssl/src/ssl_record.erl
+++ b/lib/ssl/src/ssl_record.erl
@@ -42,7 +42,8 @@
 
 %% Encoding records
 -export([encode_handshake/3, encode_alert_record/3,
-	 encode_change_cipher_spec/2, encode_data/3]).
+	 encode_change_cipher_spec/2, encode_data/3,
+   encode_heartbeat/5]).
 
 %% Compression
 -export([compress/3, uncompress/3, compressions/0]).
@@ -327,6 +328,17 @@ encode_alert_record(#alert{level = Level, description = Description},
 encode_change_cipher_spec(Version, ConnectionStates) ->
     encode_plain_text(?CHANGE_CIPHER_SPEC, Version, <<1:8>>, ConnectionStates).
 
+-spec encode_heartbeat(atom(), integer(), binary(), tls_version(), #connection_states{}) ->
+              {iolist(), #connection_states{}}.
+encode_heartbeat(TypeAtom, PayloadLen, Payload, Version, ConnectionStates) ->
+    Type = case TypeAtom of
+        request -> 1;
+        response -> 2;
+        _ -> 0
+    end,
+    encode_plain_text(?HEARTBEAT, Version,
+      <<Type:8, ?UINT16(PayloadLen), Payload/binary>>, ConnectionStates).
+
 %%--------------------------------------------------------------------
 -spec encode_data(binary(), tls_version(), #connection_states{}) ->
 			 {iolist(), #connection_states{}}.
diff --git a/lib/ssl/src/ssl_record.hrl b/lib/ssl/src/ssl_record.hrl
index c17fa53..53dcfaa 100644
--- a/lib/ssl/src/ssl_record.hrl
+++ b/lib/ssl/src/ssl_record.hrl
@@ -134,6 +134,7 @@
 -define(ALERT, 21).
 -define(HANDSHAKE, 22).
 -define(APPLICATION_DATA, 23).
+-define(HEARTBEAT, 24).
 -define(MAX_PLAIN_TEXT_LENGTH, 16384).
 -define(MAX_COMPRESSED_LENGTH, (?MAX_PLAIN_TEXT_LENGTH+1024)).
 -define(MAX_CIPHER_TEXT_LENGTH, (?MAX_PLAIN_TEXT_LENGTH+2048)).
diff --git a/lib/ssl/src/tls_connection.erl b/lib/ssl/src/tls_connection.erl
index ffa04ee..cf7f767 100644
--- a/lib/ssl/src/tls_connection.erl
+++ b/lib/ssl/src/tls_connection.erl
@@ -55,6 +55,7 @@
 	 handle_normal_shutdown/3, handle_unexpected_message/3,
 	 workaround_transport_delivery_problems/2, alert_user/5, alert_user/8
 	]).
+-export([handle_own_heartbeat/4]).
 
 %% Data handling
 -export([write_application_data/3, read_application_data/2,
@@ -217,9 +218,11 @@ hello(Hello = #client_hello{client_version = ClientVersion,
 				 State#state{connection_states  = ConnectionStates,
 					     negotiated_version = Version,
 					     session = Session,
+					     peer_hello = Hello,
 					     client_ecc = {EllipticCurves, EcPointFormats}}, ?MODULE);
         #alert{} = Alert ->
-            handle_own_alert(Alert, ClientVersion, hello, State)
+            handle_own_alert(Alert, ClientVersion, hello,
+            	State#state{peer_hello = Hello})
     end;
 hello(Hello,
       #state{connection_states = ConnectionStates0,
@@ -229,10 +232,12 @@ hello(Hello,
 	     ssl_options = SslOptions} = State) ->
     case tls_handshake:hello(Hello, SslOptions, ConnectionStates0, Renegotiation) of
 	#alert{} = Alert ->
-	    handle_own_alert(Alert, ReqVersion, hello, State);
+	    handle_own_alert(Alert, ReqVersion, hello,
+	    	State#state{peer_hello = Hello});
 	{Version, NewId, ConnectionStates, NextProtocol} ->
 	    ssl_connection:handle_session(Hello,
-					  Version, NewId, ConnectionStates, NextProtocol, State)
+					  Version, NewId, ConnectionStates, NextProtocol,
+					  State#state{peer_hello = Hello})
     end;
 
 hello(Msg, State) ->
@@ -312,6 +317,11 @@ handle_info({Protocol, _, Data}, StateName,
 	    {stop, {shutdown, own_alert}, State0}
     end;
 
+handle_info(heartbeat_timeout, StateName, State = #state{heartbeat_from = From}) ->
+	gen_fsm:reply(From, {error, timeout}),
+	{next_state, StateName, State#state{
+    		heartbeat_from = undefined, heartbeat_tref = undefined}};
+
 handle_info({CloseTag, Socket}, StateName,
             #state{socket = Socket, close_tag = CloseTag,
 		   negotiated_version = Version} = State) ->
@@ -424,6 +434,14 @@ next_state(_,Next, #ssl_tls{type = ?ALERT, fragment = EncAlerts}, State) ->
     Alerts = decode_alerts(EncAlerts),
     handle_alerts(Alerts,  {next_state, Next, State, get_timeout(State)});
 
+next_state(_, Next, #ssl_tls{type = ?HEARTBEAT, fragment = EncHeartbeat}, State) ->
+	<<Type:8, ?UINT16(PayloadLen), Rest/binary>> = EncHeartbeat,
+	TypeAtom = case Type of
+		1 -> request; 2 -> response; _ -> unknown
+	end,
+	handle_heartbeat(TypeAtom, PayloadLen, Rest,
+		{next_state, Next, State, get_timeout(State)});
+
 next_state(Current, Next, #ssl_tls{type = ?HANDSHAKE, fragment = Data},
 	   State0 = #state{protocol_buffers =
 			       #protocol_buffers{tls_handshake_buffer = Buf0} = Buffers,
@@ -823,6 +841,29 @@ renegotiate(#state{role = server,
 					       tls_handshake_history = Hs0}),
     next_state(connection, hello, Record, State#state{allow_renegotiate = true}).
 
+handle_heartbeat(request, PayloadLen, Payload,
+               {next_state, StateName, State, Timeout} = Cont) ->
+	case Payload of
+		<<RealPayload:PayloadLen/binary, _/binary>> ->
+			{ok, CS2} = handle_own_heartbeat(response, PayloadLen, RealPayload, State),
+			{next_state, StateName, State#state{connection_states = CS2}, Timeout};
+		_ ->
+			Cont
+	end;
+handle_heartbeat(response, PayloadLen, Payload,
+               {next_state, StateName, State, Timeout}) ->
+	case State of
+		#state{heartbeat_from = From, heartbeat_tref = Ref} when
+				(From =/= undefined) and (Ref =/= undefined) ->
+			{ok, cancel} = timer:cancel(Ref),
+			gen_fsm:reply(From, {ok, PayloadLen, Payload});
+		_ -> ok
+	end,
+	{next_state, StateName, State#state{heartbeat_from = undefined,
+		heartbeat_tref = undefined}, Timeout};
+handle_heartbeat(_, _, _, {next_state, _StateName, State, _Timeout}) ->
+	{stop, normal, State}.
+
 handle_alerts([], Result) ->
     Result;
 handle_alerts(_, {stop, _, _} = Stop) ->
@@ -918,6 +959,23 @@ handle_own_alert(Alert, Version, StateName,
     end,
     {stop, {shutdown, own_alert}, State}.
 
+
+handle_own_heartbeat(Type, PayloadLen, Payload,
+               #state{transport_cb = Transport,
+                       socket = Socket,
+                       negotiated_version = Version,
+                       connection_states = ConnectionStates}) ->
+	Pad = crypto:rand_bytes(16),
+	PaddedPayload = <<Payload/binary, Pad/binary>>,
+	try
+		{BinMsg, CS2} = ssl_record:encode_heartbeat(Type, PayloadLen, PaddedPayload,
+			Version, ConnectionStates),
+		Transport:send(Socket, BinMsg),
+		{ok, CS2}
+	catch _:_ ->  %% Can crash if we are in a uninitialized state
+		{ok, ConnectionStates}
+	end.
+
 handle_normal_shutdown(Alert, _, #state{socket = Socket,
 					transport_cb = Transport,
 					start_or_recv_from = StartFrom,
diff --git a/lib/ssl/src/tls_handshake.erl b/lib/ssl/src/tls_handshake.erl
index 01abefc..f8a9916 100644
--- a/lib/ssl/src/tls_handshake.erl
+++ b/lib/ssl/src/tls_handshake.erl
@@ -52,9 +52,10 @@ client_hello(Host, Port, ConnectionStates,
     Pending = ssl_record:pending_connection_state(ConnectionStates, read),
     SecParams = Pending#connection_state.security_parameters,
     CipherSuites = ssl_handshake:available_suites(UserSuites, Version),
+    HBMode = peer_allowed_to_send,
     Extensions = ssl_handshake:client_hello_extensions(Host, Version,
-						       CipherSuites,
-						       SslOpts, ConnectionStates, Renegotiation),
+						       CipherSuites, SslOpts, HBMode,
+						       ConnectionStates, Renegotiation),
 
     Id = ssl_session:client_id({Host, Port, SslOpts}, Cache, CacheCb, OwnCert),
 
diff --git a/lib/ssl/src/tls_record.erl b/lib/ssl/src/tls_record.erl
index 8810755..7bf001a 100644
--- a/lib/ssl/src/tls_record.erl
+++ b/lib/ssl/src/tls_record.erl
@@ -86,6 +86,12 @@ get_tls_records_aux(<<?BYTE(?CHANGE_CIPHER_SPEC),?BYTE(MajVer),?BYTE(MinVer),
     get_tls_records_aux(Rest, [#ssl_tls{type = ?CHANGE_CIPHER_SPEC,
 					version = {MajVer, MinVer},
 					fragment = Data} | Acc]);
+get_tls_records_aux(<<?BYTE(?HEARTBEAT),?BYTE(MajVer),?BYTE(MinVer),
+			 ?UINT16(Length), Data:Length/binary, Rest/binary>>,
+			Acc) ->
+	get_tls_records_aux(Rest, [#ssl_tls{type = ?HEARTBEAT,
+					version = {MajVer, MinVer},
+					fragment = Data} | Acc]);
 %% Matches an ssl v2 client hello message.
 %% The server must be able to receive such messages, from clients that
 %% are willing to use ssl v3 or higher, but have ssl v2 compatibility.
